#!/usr/bin/env python
####################################################################################################
# NAME
#    <NAME> - test the financial utility library
#
# SYNOPSIS
#    <NAME>
#
# AUTHOR
#    Written by Florian Barras (florian@barras.io).
#
# COPYRIGHT
#    Copyright © 2013-2022 Florian Barras <https://barras.io>.
#    The MIT License (MIT) <https://opensource.org/licenses/MIT>.
####################################################################################################

from nfin.ts import *
from nutil.test import *

####################################################################################################
# FIN TEST CONSTANTS
####################################################################################################

_FIN_TEST_CONSTANTS_______________________________ = ''

PRECISION = 14  # decimals

TEST_COUNT = 10

####################################################################################################
# FIN TEST CLASSES
####################################################################################################

__FIN_TEST_CLASSES________________________________ = ''


class TestFin(Test):

	def test_time_series(self):
		date_to = get_date()
		date_from = date_to - 2 * RELATIVE_YEAR
		index = create_date_sequence(date_from, date_to)
		series = rename(to_series(cum_diff(create_random_int_array(-5, len(index), high=6), 0)[:-1],
		                          index=index), 'Random walk')

		test('Test the time series transformations')
		for freq in (Frequency.DAYS, Frequency.WEEKS, Frequency.MONTHS, Frequency.QUARTERS,
		             Frequency.SEMESTERS, Frequency.YEARS):
			for group in (Group.FIRST, Group.LAST):
				s = transform_series(series, freq=freq, group=group,
				                     transformation=Transformation.DIFF)
				if freq is Frequency.MONTHS:
					if group is Group.FIRST:
						test(get_first(get_index(s)), '=', get_next_month_start(date_from))
						self.assert_equals(to_stamp(get_first(get_index(s))),
						                   to_stamp(get_next_month_start(date_from)))
					else:
						test(get_first(get_index(s)), '=', get_next_month_end(date_from))
						self.assert_equals(to_stamp(get_first(get_index(s))),
						                   to_stamp(get_next_month_end(date_from)))
				test(find_nearest_freq(s), '=', freq)
				self.assert_equals(find_nearest_freq(s).value, freq.value)
				if freq is not Frequency.DAYS:
					test(find_nearest_group(s, freq=freq), '=', group)
					self.assert_equals(find_nearest_group(s, freq=freq).value, group.value)

		test('Test the time series forecast')
		forecasted_series = rename(forecast_series(series, horizon=2), 'Forecast')
		forecasted_series = forecasted_series[forecasted_series.index >= date_to]
		self.assert_equals(forecasted_series[-1], -153.4974678080443)
		fig = plot_series(concat_cols(series, forecasted_series),
		                  title='Forecasting')
		fig.show()

		test('Plot the seasonal-trend decomposition')
		decomposition = decompose_series(series)
		fig = plot_decomposition(decomposition.trend, decomposition.seasonal, decomposition.resid,
		                         name='Random walk', color='blue')
		decomposition = decompose_series(forecasted_series)
		fig = plot_decomposition(decomposition.trend, decomposition.seasonal, decomposition.resid,
		                         fig=fig, name='Forecast', color='orange')
		fig.show()


####################################################################################################
# TEST MAIN
####################################################################################################

__TEST_MAIN_______________________________________ = ''


def main():
	'''Tests the financial utility library.'''
	unittest.main()


if __name__ == '__main__':
	main()



#!/usr/bin/env python
####################################################################################################
# NAME
#    <NAME> - contain financial functions for time series
#
# SYNOPSIS
#    <NAME>
#
# AUTHOR
#    Written by Florian Barras (florian@barras.io).
#
# COPYRIGHT
#    Copyright © 2013-2022 Florian Barras <https://barras.io>.
#    The MIT License (MIT) <https://opensource.org/licenses/MIT>.
####################################################################################################

from statistics import mode

from statsmodels.tsa.api import ExponentialSmoothing
from statsmodels.tsa.seasonal import STL

from ngui.chart import *
from nutil.math import *

####################################################################################################
# TIME SERIES ENUMS
####################################################################################################

__TIME_SERIES_ENUMS_______________________________ = ''


class Transformation(Enum):
	LOG = 'log'
	DIFF = 'diff'
	RETURNS = 'returns'
	LOG_RETURNS = 'log returns'


####################################################################################################
# TIME SERIES FUNCTIONS
####################################################################################################

__TIME_SERIES_____________________________________ = ''


def get_average_duration(series, per=DAY):
	return np.diff(series.index).mean() / per


def get_freq_group(series, freq=FREQUENCY, group=GROUP):
	if is_null(freq):
		freq = find_nearest_freq(series)
	if is_null(group):
		group = find_nearest_group(series, freq=freq)
	return freq, group


##################################################

def set_freq(series, freq=FREQUENCY, group=GROUP):
	freq, group = get_freq_group(series, freq=freq, group=group)
	series.index.freq = get_freq(freq=freq, group=group)


##################################################

def find_nearest_freq(series):
	return find_nearest_freq_from_days(get_average_duration(series, per=DAY))


def find_nearest_freq_from_days(day_count):
	return DAY_COUNT_TO_FREQUENCY[nearest(FREQUENCY_TO_DAY_COUNT, day_count)]


def find_nearest_freq_from_period(period=PERIOD):
	return find_nearest_freq_from_days(get_period_days(None, period=period))


def find_nearest_group(series, freq=FREQUENCY):
	if is_null(series) or freq is Frequency.DAYS:
		return GROUP
	if freq is Frequency.WEEKS:
		if mode(get_weekdays(series, use_index=True)) < DAYS_PER_WEEK / 2:
			return Group.FIRST
	elif freq is Frequency.MONTHS:
		if mode(get_days(series, use_index=True)) < DAYS_PER_MONTH / 2:
			return Group.FIRST
	elif freq is Frequency.QUARTERS:
		if mode(get_days(series, use_index=True, year=True)) % DAYS_PER_QUARTER < DAYS_PER_QUARTER / 2:
			return Group.FIRST
	elif freq is Frequency.SEMESTERS:
		if mode(get_days(series, use_index=True, year=True)) % DAYS_PER_SEMESTER < DAYS_PER_SEMESTER / 2:
			return Group.FIRST
	elif freq is Frequency.YEARS:
		if mode(get_days(series, use_index=True, year=True)) < DAYS_PER_YEAR / 2:
			return Group.FIRST
	return Group.LAST


def find_nearest_freq_group(series):
	freq = find_nearest_freq(series)
	group = find_nearest_group(series, freq=freq)
	return freq, group


#########################

def group_series(series, clean=False, freq=FREQUENCY, sort=True):
	if clean:
		series = remove_null(series)
	if is_empty(series):
		return series
	if freq is Frequency.WEEKS:
		index = get_year_weeks(series, use_index=True)
	elif freq is Frequency.MONTHS:
		index = [get_years(series, use_index=True),
		         get_months(series, use_index=True)]
	elif freq is Frequency.QUARTERS:
		index = [get_years(series, use_index=True),
		         get_quarters(series, use_index=True)]
	elif freq is Frequency.SEMESTERS:
		index = [get_years(series, use_index=True),
		         get_semesters(series, use_index=True)]
	elif freq is Frequency.YEARS:
		index = get_years(series, use_index=True)
	else:
		index = [get_years(series, use_index=True),
		         get_months(series, use_index=True),
		         get_days(series, use_index=True)]
	return series.groupby(index, sort=sort)


def ungroup_series(series, clean=False, freq=FREQUENCY, end=True):
	if clean:
		series = remove_null(series)
	if is_empty(series):
		return series
	if freq is Frequency.WEEKS:
		series.index = ([get_end_period(y, w=w) for y, w in series.index] if end else
		                [get_start_period(y, w=w) for y, w in series.index])
	elif freq is Frequency.MONTHS:
		series.index = ([get_end_period(y, m=m) for y, m in series.index] if end else
		                [get_start_period(y, m=m) for y, m in series.index])
	elif freq is Frequency.QUARTERS:
		series.index = ([get_end_period(y, q=q) for y, q in series.index] if end else
		                [get_start_period(y, q=q) for y, q in series.index])
	elif freq is Frequency.SEMESTERS:
		series.index = ([get_end_period(y, s=s) for y, s in series.index] if end else
		                [get_start_period(y, s=s) for y, s in series.index])
	elif freq is Frequency.YEARS:
		series.index = ([get_end_period(y) for y in series.index] if end else
		                [get_start_period(y) for y in series.index])
	else:
		series.index = ([get_end_period(y, m=m, d=d) for y, m, d in series.index] if end else
		                [get_start_period(y, m=m, d=d) for y, m, d in series.index])
	return series


# • TIME SERIES TRANSFORMATION #####################################################################

__TIME_SERIES_TRANSFORMATION______________________ = ''


def get_diff(series, periods=1):
	if is_table(series):
		return remove_null(series.diff(periods=periods))
	return to_array(series[periods:]) - to_array(series[:-periods])


def cum_diff(series, offset):
	series = concat(offset, series)
	if is_table(series):
		return series.cumsum()
	return np.cumsum(series)


def get_returns(series, periods=1):
	if is_table(series):
		return remove_null(series.pct_change(periods=periods))
	return to_array(series[periods:]) / to_array(series[:-periods]) - 1


def cum_returns(series, offset):
	series = concat(offset, series + 1)
	if is_table(series):
		return series.cumprod()
	return np.cumprod(series)


def get_log_returns(series, periods=1):
	return log(get_returns(series, periods=periods) + 1)


def cum_log_returns(series, offset):
	return cum_returns(exp(series) - 1, offset)


#########################

def get_moving_average(series, window):
	return remove_null(series.rolling(window).mean())


#########################

def get_period_over_period(series, freq=FREQUENCY):
	if is_null(freq):
		return series
	s = series.copy()
	if freq is Frequency.DAYS:
		s = shift_dates(s, days=1)
	elif freq is Frequency.WEEKS:
		s = shift_dates(s, days=7)
	elif freq is Frequency.MONTHS:
		s = shift_dates(s, months=1)
	elif freq is Frequency.QUARTERS:
		s = shift_dates(s, months=3)
	elif freq is Frequency.SEMESTERS:
		s = shift_dates(s, months=6)
	elif freq is Frequency.YEARS:
		s = shift_dates(s, years=1)
	return remove_null(subtract(series, s))


##################################################

def clean_series(series, group=GROUP):
	return sort_index(unique(remove_null(series), group=group))


def prepare_series(series, date_from=None, date_to=None, fill=False, interpolate=True,
                   freq=FREQUENCY, group=GROUP):
	freq, group = get_freq_group(series, freq=freq, group=group)
	series = transform_series(series, freq=freq, group=group)
	if is_null(date_from):
		date_from = get_first(series.index)
	if is_null(date_to):
		date_to = get_last(series.index)
	series = fill_null_rows(series[(series.index >= date_from - FREQUENCY_TO_RELATIVE_DURATION[freq]) &
	                               (series.index <= date_to)],
	                        create_datetime_sequence(date_from, date_to, freq=freq, group=group))
	set_freq(series, freq=freq, group=group)
	if fill:
		series = series.fillna(method='ffill')
	elif interpolate:
		series = series.interpolate()
	return series[series.index >= date_from]


#########################

def decompose_series(series, seasonal_period=1, freq=FREQUENCY, group=GROUP):
	'''Decomposes the specified time series into trend and seasonality using the seasonal-trend
	decomposition procedure STL based on LOESS of R. B. Cleveland, W. S. Cleveland, J.E. McRae, and
	I. Terpenning (1990).'''
	freq, group = get_freq_group(series, freq=freq, group=group)
	series = prepare_series(series, freq=freq, group=group)
	seasonal_period_length = seasonal_period * get_period_length(get_date(), freq=freq)
	return STL(series, period=seasonal_period_length).fit()


#########################

def forecast_series(series, horizon=1, initialization_method='estimated', trend='add',
                    seasonal='add', seasonal_period=1, freq=FREQUENCY, group=GROUP):
	'''Forecasts the specified time series using Holt Winter's Exponential Smoothing (2014).'''
	freq, group = get_freq_group(series, freq=freq, group=group)
	series = prepare_series(series, freq=freq, group=group)
	seasonal_period_length = seasonal_period * get_period_length(get_date(), freq=freq)
	predictions = to_frame([])
	for s in to_series(series) if is_frame(series) else [series]:
		model = ExponentialSmoothing(s, initialization_method=initialization_method, trend=trend,
		                             seasonal=seasonal, seasonal_periods=seasonal_period_length).fit()
		prediction = set_names(model.forecast(steps=horizon * seasonal_period_length), s)
		predictions = concat_cols(predictions, concat_rows(s, prediction))
	return predictions


#########################

def transform_series(series, clean=True, method=None, sort=True, freq=FREQUENCY, group=GROUP,
                     transformation=None):
	if clean:
		series = remove_null(series)
	if sort:
		series = sort_index(series)
	if is_empty(series):
		return series
	freq, group = get_freq_group(series, freq=freq, group=group)
	series = group_series(series, freq=freq)
	if group is Group.COUNT:
		series = series.count()
	elif group is Group.FIRST:
		series = series.first()
	elif group is Group.LAST:
		series = series.last()
	elif group is Group.MIN:
		series = series.min()
	elif group is Group.MAX:
		series = series.max()
	elif group is Group.MEAN:
		series = series.mean()
	elif group is Group.MEDIAN:
		series = series.median()
	elif group is Group.STD:
		series = series.std()
	elif group is Group.VAR:
		series = series.var()
	elif group is Group.SUM:
		series = series.sum()
	series = ungroup_series(series, clean=clean, freq=freq, end=group != Group.FIRST)
	if not is_null(method):
		series = series.asfreq(get_freq(freq, group), method=method)
	if transformation is Transformation.LOG:
		series = log(series)
	elif transformation is Transformation.DIFF:
		series = get_diff(series)
	elif transformation is Transformation.RETURNS:
		series = get_returns(series)
	elif transformation is Transformation.LOG_RETURNS:
		series = get_log_returns(series)
	return series


def untransform_series(series, offset, clean=True, transformation=None):
	if clean:
		series = remove_null(series)
	if is_empty(series):
		return series
	if transformation is Transformation.LOG:
		return exp(series)
	elif transformation is Transformation.DIFF:
		return cum_diff(series, offset)
	elif transformation is Transformation.RETURNS:
		return cum_returns(series, offset)
	elif transformation is Transformation.LOG_RETURNS:
		return cum_log_returns(series, offset)
	return series


# • TIME SERIES FIGURE #############################################################################

__TIME_SERIES_FIGURE______________________________ = ''


def plot_decomposition(trend, seasonal, residual,
                       # Figure
                       fig=None, title='Seasonal-Trend Decomposition', title_x='Time', title_y=None,
                       width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None,
                       # Chart
                       color='black', trend_color='red', seasonal_color='gray',
                       residual_color='lightgray', line_width=DEFAULT_LINE_WIDTH,
                       marker_size=DEFAULT_MARKER_SIZE, name=None, stackgroup=None, yaxis=0,
                       # Flags
                       show_legend=False):
	if is_null(fig):
		fig = create_figure(title=title, title_x=title_x, title_y=title_y,
		                    width=width, height=height, margin=margin)
	if is_null(stackgroup):
		stackgroup = generate_string(10)
	fig.add_trace(draw(x=trend.index, y=get_col(trend),
	                   # Chart
	                   color=trend_color, fill='none', line_width=line_width,
	                   marker_size=marker_size, name=paste(name, '(Trend Component)'),
	                   stackgroup=stackgroup, yaxis=yaxis,
	                   # Flags
	                   show_legend=show_legend))
	fig.add_trace(draw(x=seasonal.index, y=get_col(seasonal),
	                   # Chart
	                   color=seasonal_color, fill='tonexty', line_width=line_width,
	                   marker_size=marker_size, name=paste(name, '(Seasonal Component)'),
	                   stackgroup=stackgroup, yaxis=yaxis,
	                   # Flags
	                   show_legend=show_legend))
	fig.add_trace(draw(x=residual.index, y=get_col(residual),
	                   # Chart
	                   color=residual_color, fill='tonexty', line_width=line_width,
	                   marker_size=marker_size, name=paste(name, '(Residual Component)'),
	                   stackgroup=stackgroup, yaxis=yaxis,
	                   # Flags
	                   show_legend=show_legend))
	series = trend + seasonal + residual
	fig.add_trace(draw(x=series.index, y=get_col(series),
	                   # Chart
	                   color=color, line_width=line_width, marker_size=marker_size, name=name,
	                   yaxis=yaxis))
	return fig



#!/usr/bin/env python
####################################################################################################
# NAME
#    <NAME> - contain the build script for setuptools
#
# SYNOPSIS
#    <NAME>
#
# AUTHOR
#    Written by Florian Barras (florian@barras.io).
#
# COPYRIGHT
#    Copyright © 2013-2022 Florian Barras <https://barras.io>.
#    The MIT License (MIT) <https://opensource.org/licenses/MIT>.
####################################################################################################

from os import path

from setuptools import find_packages, setup

####################################################################################################
# SETUP CONSTANTS
####################################################################################################

__SETUP_CONSTANTS_________________________________ = ''

NAME = 'nfin'
VERSION = '1.0.0.post137'
DESCRIPTION = 'Financial utility library'
DIR = path.abspath(path.dirname(__file__))
with open(path.join(DIR, 'README.md'), encoding='utf-8') as f:
	LONG_DESCRIPTION = f.read()

AUTHOR = 'Florian Barras'
AUTHOR_EMAIL = 'florian@barras.io'
LICENSE = 'MIT'
LICENSE_FILES = ['LICENSE']
URL = 'https://github.com/b-io/io.barras/tree/master/python/neptune/nfin'

PACKAGES = [
	'ngui', 'nutil', 'statsmodels'
]

####################################################################################################
# SETUP
####################################################################################################

__SETUP___________________________________________ = ''

setup(
	# The project name.
	#
	# Note that there are some restrictions on what makes a valid project name:
	# https://packaging.python.org/specifications/core-metadata/#name
	name=NAME,  # Required

	# The project version.
	#
	# Note that the version should comply with PEP 440:
	# https://www.python.org/dev/peps/pep-0440/
	# For a discussion on single-sourcing the version across setup.py and the project code, see:
	# https://packaging.python.org/en/latest/single_source_version.html
	version=VERSION,  # Required

	# The project one-line description (tagline).
	#
	# Note that this corresponds to the "Summary" metadata field:
	# https://packaging.python.org/specifications/core-metadata/#summary
	description=DESCRIPTION,  # Optional

	# The project long description (README).
	#
	# Note that this corresponds to the "Description" metadata field:
	# https://packaging.python.org/specifications/core-metadata/#description-optional
	long_description=LONG_DESCRIPTION,  # Optional

	# The content type of the project long description (either "text/plain", "text/x-rst" or
	# "text/markdown").
	#
	# Note that this corresponds to the "Description-Content-Type" metadata field:
	# https://packaging.python.org/specifications/core-metadata/#description-content-type-optional
	long_description_content_type='text/markdown',  # Optional

	# The project author.
	author=AUTHOR,  # Optional

	# The email address of the project author.
	author_email=AUTHOR_EMAIL,  # Optional

	# The email address of the project author.
	license=LICENSE,  # Optional

	# The email address of the project author.
	license_files=LICENSE_FILES,  # Optional

	# The project main homepage.
	#
	# Note that this corresponds to the "Home-Page" metadata field:
	# https://packaging.python.org/specifications/core-metadata/#home-page-optional
	url=URL,  # Optional

	# The project classifiers.
	#
	# For a list of valid classifiers, see:
	# https://pypi.org/classifiers/
	classifiers=[  # Optional
		# The project maturity
		'Development Status :: 4 - Beta',

		# The project audience
		'Intended Audience :: Developers',

		# The project license
		'License :: OSI Approved :: MIT License',

		# The project platform
		'Operating System :: OS Independent',

		# The Python versions supported by the project
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.8',
		'Programming Language :: Python :: 3.9',
		'Programming Language :: Python :: 3.10',
		'Programming Language :: Python :: Implementation :: CPython',
		'Programming Language :: Python :: Implementation :: PyPy',

		# The project topics
		'Topic :: Scientific/Engineering :: Visualization',
		'Topic :: Utilities'
	],

	# The project keywords.
	keywords=[
		'finance',
		'time series',
		'utilities'
	],  # Optional

	# The path to the project packages.
	package_dir={'': 'source'},  # Optional

	# The project package directories (use find_packages()).
	#
	# Note that, alternatively, to distribute a single Python file, the "py_modules" argument can be
	# used instead as follows, which will expect a file called "modules.py" to exist:
	# py_modules=['modules'],
	packages=find_packages(),  # Required

	# The Python versions supported by the project.
	python_requires='>=3.8',  # Required

	# The project dependencies.
	install_requires=PACKAGES,  # Optional

	# The environment-specific project dependencies.
	#
	# Note that the users will be able to install these additional dependencies using the "extras"
	# syntax as follows:
	#   $ pip install nfin[dev]
	extras_require={  # Optional
		'dev': ['check-manifest'],
		'test': ['coverage']
	},

	# The list of project resources (included in the packages) that need to be installed.
	package_data={  # Optional
		'common': ['resources/common.properties']
	},

	# The list of project resources (not included in the packages) that need to be installed.
	#
	# Note that although "package_data" is the preferred approach, in some case you may need to
	# place resources outside of the packages. See:
	# http://docs.python.org/distutils/setupscript.html#installing-additional-files
	data_files=[  # Optional
	],

	# The platform-specific project entry points (in preference to the "scripts" keyword).
	#
	# Note that these entry points provide cross-platform support and allow "pip" to create the
	# appropriate form of executable for the target platform.
	entry_points={  # Optional
		'console_scripts': [
			'main=nfin:main'
		],
	},

	# The list of project URLs.
	#
	# Note that this corresponds to the "Project-URL" metadata fields:
	# https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
	project_urls={  # Optional
		'Bug Tracker': 'https://github.com/b-io/io.barras/issues',
		'Documentation': 'https://repo.barras.io',
		'Source Code': URL
	}
)
